

#include "ScreenBoard.H"
#include "gba.H"
#include "fonts.H"
#include "tiles.H"
#include "palette.H"
#include "GameStatus.H"


using namespace std;
using namespace avelino;


const ScreenBoard::Direction ScreenBoard::DIRECTIONS[NUM_DIRECTIONS] = {
	{0, -1}, {0, 1}, {-1, 0}, {1, 0}, {-1, -1}, {1, -1}, {-1, 1}, {1, 1},
	{0, 0}    // DIRECTION_SELF
};


bool ScreenBoard::isBoardNeighbourPassable(const Level &level, int16_t x, int16_t y, const Direction &d, ConsiderTheBlocks considerTheBlocks) {
	int16_t neighbourX = x + d.x;
	int16_t neighbourY = y + d.y;
	if ((neighbourX < 0) || (neighbourX >= Level::BOARD_WIDTH) || (neighbourY < 0) || (neighbourY >= Level::BOARD_HEIGHT))
		return false;
	else {
		uint16_t t = level.boardTiles[neighbourX + (neighbourY * Level::BOARD_WIDTH)];
		if (considerTheBlocks == ConsiderTheBlocks::YES)
			return (((t & 0x80) == 0) && ((t & 0x07) == 0));
		else
			return ((t & 0x80) == 0);
	}
}


const uint16_t ScreenBoard::TILE_0_MAP[8] = {
	                                // lP, tlP, tP
	BOARD_GRASS_1_TILE_INDEX,       //  0   0   0
	BOARD_BORDER_B_TILE_INDEX,      //  0   0   1
	BOARD_BORDER_BR_CC_TILE_INDEX,  //  0   1   0
	BOARD_BORDER_B_TILE_INDEX,      //  0   1   1
	BOARD_BORDER_R_TILE_INDEX,      //  1   0   0
	BOARD_BORDER_BR_CV_TILE_INDEX,  //  1   0   1
	BOARD_BORDER_R_TILE_INDEX,      //  1   1   0
	BOARD_BORDER_BR_CV_TILE_INDEX   //  1   1   1
};


const uint16_t ScreenBoard::TILE_1_MAP[8] = {
	                                // tP, trP, rP
	BOARD_GRASS_1_TILE_INDEX,       //  0   0   0
	BOARD_BORDER_L_TILE_INDEX,      //  0   0   1
	BOARD_BORDER_BL_CC_TILE_INDEX,  //  0   1   0
	BOARD_BORDER_L_TILE_INDEX,      //  0   1   1
	BOARD_BORDER_B_TILE_INDEX,      //  1   0   0
	BOARD_BORDER_BL_CV_TILE_INDEX,  //  1   0   1
	BOARD_BORDER_B_TILE_INDEX,      //  1   1   0
	BOARD_BORDER_BL_CV_TILE_INDEX   //  1   1   1
};


const uint16_t ScreenBoard::TILE_2_MAP[8] = {
	                                // bP, blP, lP
	BOARD_GRASS_1_TILE_INDEX,       //  0   0   0
    BOARD_BORDER_R_TILE_INDEX,      //  0   0   1
    BOARD_BORDER_TR_CC_TILE_INDEX,  //  0   1   0
    BOARD_BORDER_R_TILE_INDEX,      //  0   1   1
    BOARD_BORDER_T_TILE_INDEX,      //  1   0   0
    BOARD_BORDER_TR_CV_TILE_INDEX,  //  1   0   1
    BOARD_BORDER_T_TILE_INDEX,      //  1   1   0
    BOARD_BORDER_TR_CV_TILE_INDEX   //  1   1   1
};


const uint16_t ScreenBoard::TILE_3_MAP[8] = {
	                                // rP, brP, bP
	BOARD_GRASS_1_TILE_INDEX,       //  0   0   0
	BOARD_BORDER_T_TILE_INDEX,      //  0   0   1
	BOARD_BORDER_TL_CC_TILE_INDEX,  //  0   1   0
	BOARD_BORDER_T_TILE_INDEX,      //  0   1   1
	BOARD_BORDER_L_TILE_INDEX,      //  1   0   0
	BOARD_BORDER_TL_CV_TILE_INDEX,  //  1   0   1
	BOARD_BORDER_L_TILE_INDEX,      //  1   1   0
	BOARD_BORDER_TL_CV_TILE_INDEX   //  1   1   1
};


uint16_t ScreenBoard::getTile(TileIndex tileIndex, bool tlPassable, bool tPassable, bool trPassable, bool rPassable, bool brPassable, bool bPassable, bool blPassable, bool lPassable) {
	uint16_t i = 0;
	const uint16_t *tileMap;
	if (tileIndex == TileIndex::TILE_0) {
		i = ((lPassable) ? 4 : 0) | ((tlPassable) ? 2 : 0) | ((tPassable) ? 1 : 0);
		tileMap = TILE_0_MAP;
	}
	else if (tileIndex == TileIndex::TILE_1) {
		i = ((tPassable) ? 4 : 0) | ((trPassable) ? 2 : 0) | ((rPassable) ? 1 : 0);
		tileMap = TILE_1_MAP;
	}
	else if (tileIndex == TileIndex::TILE_2) {
		i = ((bPassable) ? 4 : 0) | ((blPassable) ? 2 : 0) | ((lPassable) ? 1 : 0);
		tileMap = TILE_2_MAP;
	}
	else if (tileIndex == TileIndex::TILE_3) {
		i = ((rPassable) ? 4 : 0) | ((brPassable) ? 2 : 0) | ((bPassable) ? 1 : 0);
		tileMap = TILE_3_MAP;
	}
	return tileMap[i];
}


void ScreenBoard::drawFinish(int16_t x, int16_t y, uint16_t color, uint16_t *destBG) {
	color = (color - 1) & 3;    // 1..4 --> 0..3
	uint16_t tileBase = (color << 2) + BOARD_RED_FINISH_0_TILE_INDEX + BOARD_TILES_OFFSET;
	x = 1 + (x << 1);    // convert board coordinates into screen coordinates
	y = 1 + (y << 1);
	destBG[(y << 5) + x] = tileBase;
	destBG[(y << 5) + (x + 1)] = tileBase + 1;
	destBG[((y + 1) << 5) + x] = tileBase + 2;
	destBG[((y + 1) << 5) + (x + 1)] = tileBase + 3;
}


void ScreenBoard::drawBoard(const Level &level, uint16_t *destBG, uint16_t *destBGForEndings) {
	int16_t tileY = 1;
	for (int16_t y = 0; y < Level::BOARD_HEIGHT; y++, tileY += 2) {
		int16_t tileX = 1;
		for (int16_t x = 0; x < Level::BOARD_WIDTH; x++, tileX += 2) {
			if (this->isBoardNeighbourPassable(level, x, y, DIRECTIONS[DIRECTION_SELF])) {
				// passable, check if it's a finish board place
				uint16_t v = level.boardTiles[x + (y * Level::BOARD_WIDTH)];
				if (v & 0x0070)
					this->drawFinish(x, y, (v & 0x0070) >> 4, destBGForEndings);
			}
			else {
				// non passable so calculate the "passability" of the eight neighbours to obtain the correct border tiles to paint
				bool tlPassable = this->isBoardNeighbourPassable(level, x, y, DIRECTIONS[DIRECTION_TOP_LEFT]);
				bool tPassable = this->isBoardNeighbourPassable(level, x, y, DIRECTIONS[DIRECTION_TOP]);
				bool trPassable = this->isBoardNeighbourPassable(level, x, y, DIRECTIONS[DIRECTION_TOP_RIGHT]);
				bool rPassable = this->isBoardNeighbourPassable(level, x, y, DIRECTIONS[DIRECTION_RIGHT]);
				bool brPassable = this->isBoardNeighbourPassable(level, x, y, DIRECTIONS[DIRECTION_BOTTOM_RIGHT]);
				bool bPassable = this->isBoardNeighbourPassable(level, x, y, DIRECTIONS[DIRECTION_BOTTOM]);
				bool blPassable = this->isBoardNeighbourPassable(level, x, y, DIRECTIONS[DIRECTION_BOTTOM_LEFT]);
				bool lPassable = this->isBoardNeighbourPassable(level, x, y, DIRECTIONS[DIRECTION_LEFT]);
				// put the corresponding four tiles
				uint16_t tile0 = this->getTile(TileIndex::TILE_0, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
				destBG[(tileY << 5) + tileX] = tile0 + BOARD_TILES_OFFSET;
				uint16_t tile1 = this->getTile(TileIndex::TILE_1, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
				destBG[(tileY << 5) + (tileX + 1)] = tile1 + BOARD_TILES_OFFSET;
				uint16_t tile2 = this->getTile(TileIndex::TILE_2, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
				destBG[((tileY + 1) << 5) + tileX] = tile2 + BOARD_TILES_OFFSET;
				uint16_t tile3 = this->getTile(TileIndex::TILE_3, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
				destBG[((tileY + 1) << 5) + (tileX + 1)] = tile3 + BOARD_TILES_OFFSET;
			}
		}
	}
}


void ScreenBoard::drawBoardBorder2(const Level &level, uint16_t *destBG) {
	uint16_t *p = destBG;
	uint16_t tile0, tile1, tile2, tile3;
	bool tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable;
	// top left border tile
	tlPassable = tPassable = trPassable = rPassable = bPassable = blPassable = lPassable = false;
	brPassable = this->isBoardNeighbourPassable(level, -1, -1, DIRECTIONS[DIRECTION_BOTTOM_RIGHT]);
	tile3 = this->getTile(TileIndex::TILE_3, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
	*p = tile3 + BOARD_TILES_OFFSET;
	p++;
	// top tiles
	for (int16_t x = 0; x < Level::BOARD_WIDTH; x++) {
		tlPassable = tPassable = trPassable = rPassable = lPassable = false;
		brPassable = this->isBoardNeighbourPassable(level, x, -1, DIRECTIONS[DIRECTION_BOTTOM_RIGHT]);
		bPassable = this->isBoardNeighbourPassable(level, x, -1, DIRECTIONS[DIRECTION_BOTTOM]);
		blPassable = this->isBoardNeighbourPassable(level, x, -1, DIRECTIONS[DIRECTION_BOTTOM_LEFT]);
		tile2 = this->getTile(TileIndex::TILE_2, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
		*p = tile2 + BOARD_TILES_OFFSET;
		p++;
		tile3 = this->getTile(TileIndex::TILE_3, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
		*p = tile3 + BOARD_TILES_OFFSET;
		p++;
	}
	// top right border
	tlPassable = tPassable = trPassable = rPassable = brPassable = bPassable = lPassable = false;
	blPassable = this->isBoardNeighbourPassable(level, Level::BOARD_WIDTH, -1, DIRECTIONS[DIRECTION_BOTTOM_LEFT]);
	tile2 = this->getTile(TileIndex::TILE_2, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
	*p = tile2 + BOARD_TILES_OFFSET;
	p++;
	// fill until end of background
	memset16(p, BOARD_TILES_OFFSET + BOARD_GRASS_1_TILE_INDEX, 32 - (Level::BOARD_WIDTH << 1) - 2);
	p += 32 - (Level::BOARD_WIDTH << 1) - 2;
	// left and right border tiles
	for (int16_t y = 0; y < Level::BOARD_HEIGHT; y++) {
		tlPassable = tPassable = lPassable = blPassable = bPassable = false;
		trPassable = this->isBoardNeighbourPassable(level, -1, y, DIRECTIONS[DIRECTION_TOP_RIGHT]);
		rPassable = this->isBoardNeighbourPassable(level, -1, y, DIRECTIONS[DIRECTION_RIGHT]);
		brPassable = this->isBoardNeighbourPassable(level, -1, y, DIRECTIONS[DIRECTION_BOTTOM_RIGHT]);
		tile1 = this->getTile(TileIndex::TILE_1, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
		tile3 = this->getTile(TileIndex::TILE_3, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
		*p = tile1 + BOARD_TILES_OFFSET;
		*(p + 32) = tile3 + BOARD_TILES_OFFSET;
		p += 19;
		tPassable = trPassable = rPassable = brPassable = bPassable = false;
		tlPassable = this->isBoardNeighbourPassable(level, Level::BOARD_HEIGHT, y, DIRECTIONS[DIRECTION_TOP_LEFT]);
		lPassable = this->isBoardNeighbourPassable(level, Level::BOARD_HEIGHT, y, DIRECTIONS[DIRECTION_LEFT]);
		blPassable = this->isBoardNeighbourPassable(level, Level::BOARD_HEIGHT, y, DIRECTIONS[DIRECTION_BOTTOM_LEFT]);
		tile0 = this->getTile(TileIndex::TILE_0, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
		tile2 = this->getTile(TileIndex::TILE_2, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
		*p = tile0 + BOARD_TILES_OFFSET;
		*(p + 32) = tile2 + BOARD_TILES_OFFSET;
		p++;
		// fill until end of background
		memset16(p, BOARD_TILES_OFFSET + BOARD_GRASS_1_TILE_INDEX, 32 - (Level::BOARD_WIDTH << 1) - 2);
		memset16(p + 32, BOARD_TILES_OFFSET + BOARD_GRASS_1_TILE_INDEX, 32 - (Level::BOARD_WIDTH << 1) - 2);
		p += 64 - (Level::BOARD_WIDTH << 1) - 2;
	}
	// bottom left border tile
	tlPassable = tPassable = rPassable = brPassable = bPassable = blPassable = lPassable = false;
	trPassable = this->isBoardNeighbourPassable(level, -1, Level::BOARD_HEIGHT, DIRECTIONS[DIRECTION_TOP_RIGHT]);
	tile1 = this->getTile(TileIndex::TILE_1, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
	*p = tile1 + BOARD_TILES_OFFSET;
	p++;
	// bottom tiles
	for (int16_t x = 0; x < Level::BOARD_WIDTH; x++) {
		rPassable = brPassable = bPassable = blPassable = lPassable = false;
		tlPassable = this->isBoardNeighbourPassable(level, x, Level::BOARD_HEIGHT, DIRECTIONS[DIRECTION_TOP_LEFT]);
		tPassable = this->isBoardNeighbourPassable(level, x, Level::BOARD_HEIGHT, DIRECTIONS[DIRECTION_TOP]);
		trPassable = this->isBoardNeighbourPassable(level, x, Level::BOARD_HEIGHT, DIRECTIONS[DIRECTION_TOP_RIGHT]);
		tile0 = this->getTile(TileIndex::TILE_0, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
		*p = tile0 + BOARD_TILES_OFFSET;
		p++;
		tile1 = this->getTile(TileIndex::TILE_1, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
		*p = tile1 + BOARD_TILES_OFFSET;
		p++;
	}
	// bottom right border
	tPassable = trPassable = rPassable = brPassable = bPassable = blPassable = lPassable = false;
	tlPassable = this->isBoardNeighbourPassable(level, Level::BOARD_WIDTH, Level::BOARD_HEIGHT, DIRECTIONS[DIRECTION_TOP_LEFT]);
	tile0 = this->getTile(TileIndex::TILE_0, tlPassable, tPassable, trPassable, rPassable, brPassable, bPassable, blPassable, lPassable);
	*p = tile0 + BOARD_TILES_OFFSET;
	p++;
	// fill until end of background
	memset16(p, BOARD_TILES_OFFSET + BOARD_GRASS_1_TILE_INDEX, 32 - (Level::BOARD_WIDTH << 1) - 2);
	p += 32 - (Level::BOARD_WIDTH << 1) - 2;
}


void ScreenBoard::drawBoardBorder(const Level &level, uint16_t *destBG) {
	// by now put pure grass on border
	const uint16_t WIDTH = 20;
	for (uint16_t i = 0; i < 32; i++, destBG++)
		*destBG = BOARD_TILES_OFFSET + BOARD_GRASS_1_TILE_INDEX;
	for (uint16_t row = 0; row < 18; row++) {
		for (uint16_t i = 0; i < 32; i++, destBG++) {
			if ((i < 1) || (i > 18))
				*destBG = BOARD_TILES_OFFSET + BOARD_GRASS_1_TILE_INDEX;
		}
	}
	for (uint16_t i = 0; i < 32; i++, destBG++)
		*destBG = BOARD_TILES_OFFSET + BOARD_GRASS_1_TILE_INDEX;
}


void ScreenBoard::locateBlock(uint16_t index, int16_t x, int16_t y) {
	this->blockLocations[index].x = x;
	this->blockLocations[index].y = y;
	uint16_t regOffset = 20 + (index << 2);
	x = (x << 4) + 8;
	y = (y << 4) + 8;
	OAM[regOffset]     = (y & 0x00FF) | (((uint16_t) 1) << 13);    // 256 colors, square
	OAM[regOffset + 1] = (x & 0x01FF) | (((uint16_t) 1) << 14);    // size 16x16
}


void ScreenBoard::hideBlock(uint16_t index) {
	uint16_t regOffset = 20 + (index << 2);
	OAM[regOffset] |= (((uint16_t) 1) << 9);
}


void ScreenBoard::locateBlocks(Level &level, uint16_t &firstAvailableBlockIndex) {
	this->hideSprites();
	for (uint16_t i = 0; i < NUM_BLOCKS; i++) {
		this->blockEnabled[i] = false;
		this->hideBlock(i);
	}
	bool assignedFirstAvailableBlockIndex = false;
	uint16_t *pBoardTiles = level.boardTiles;
	for (int16_t y = 0; y < Level::BOARD_HEIGHT; y++) {
		for (int16_t x = 0; x < Level::BOARD_WIDTH; x++, pBoardTiles++) {
			uint16_t block = *pBoardTiles & 0x0007;
			if (block != 0) {
				if (!assignedFirstAvailableBlockIndex) {
					firstAvailableBlockIndex = block - 1;
					assignedFirstAvailableBlockIndex = true;
				}
				this->blockEnabled[block - 1] = true;
				this->locateBlock(block - 1, x, y);
			}
		}
	}
}


void ScreenBoard::updateMarkerFromIndex() {
	int16_t x = this->blockLocations[this->markerIndex].x;
	int16_t y = this->blockLocations[this->markerIndex].y;
	x = (x << 4) + 8;
	y = (y << 4) + 8;
	OAM[0] = (OAM[0] & (((uint16_t) 1) << 9)) | (y & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square, backup disable bit to keep the tilt effect
	OAM[1] = (x & 0x01FF) | (((uint16_t) 1) << 14);    // size 16x16
}


void ScreenBoard::startMoveBlockOnMarker(Movement m) {
	GameStatus *gs = static_cast<GameStatus *>(this->interScreenData);
	if (this->markerIndex >= NUM_BLOCKS)
		return;
	if (!this->blockEnabled[this->markerIndex])
		return;
	// hide marker
	OAM[0] = OAM[0] | (((uint16_t) 1) << 9);
	// hide arrows
	this->hideArrows();
	// get board coordinates of the block to move (under marker)
	int16_t xBoard = this->blockLocations[this->markerIndex].x;
	int16_t yBoard = this->blockLocations[this->markerIndex].y;
	this->xBoardDest = xBoard;
	this->yBoardDest = yBoard;
	this->xSpeedFP = 0;
	this->ySpeedFP = 0;
	if (m == Movement::UP) {
		do {
			this->yBoardDest--;
		} while (this->isBoardNeighbourPassable(gs->currentLevel, this->xBoardDest, this->yBoardDest, DIRECTIONS[DIRECTION_SELF], ConsiderTheBlocks::YES));
		this->yBoardDest++;
		this->ySpeedFP = -BLOCK_SPEED_FP;
	}
	else if (m == Movement::DOWN) {
		do {
			this->yBoardDest++;
		} while (this->isBoardNeighbourPassable(gs->currentLevel, this->xBoardDest, this->yBoardDest, DIRECTIONS[DIRECTION_SELF], ConsiderTheBlocks::YES));
		this->yBoardDest--;
		this->ySpeedFP = BLOCK_SPEED_FP;
	}
	else if (m == Movement::LEFT) {
		do {
			this->xBoardDest--;
		} while (this->isBoardNeighbourPassable(gs->currentLevel, this->xBoardDest, this->yBoardDest, DIRECTIONS[DIRECTION_SELF], ConsiderTheBlocks::YES));
		this->xBoardDest++;
		this->xSpeedFP = -BLOCK_SPEED_FP;
	}
	else if (m == Movement::RIGHT) {
		do {
			this->xBoardDest++;
		} while (this->isBoardNeighbourPassable(gs->currentLevel, this->xBoardDest, this->yBoardDest, DIRECTIONS[DIRECTION_SELF], ConsiderTheBlocks::YES));
		this->xBoardDest--;
		this->xSpeedFP = BLOCK_SPEED_FP;
	}
	this->xFP = ((int32_t) (8 + (xBoard << 4))) << 16;
	this->yFP = ((int32_t) (8 + (yBoard << 4))) << 16;
	this->xBoardSrc = xBoard;
	this->yBoardSrc = yBoard;
	this->xDestFP = ((int32_t) (8 + (this->xBoardDest << 4))) << 16;
	this->yDestFP = ((int32_t) (8 + (this->yBoardDest << 4))) << 16;
}


void ScreenBoard::updateMoveBlockOnMarker(bool &finished) {
	GameStatus *gs = static_cast<GameStatus *>(this->interScreenData);
	uint16_t regOffset = 20;
	if (this->markerIndex == 0)           // red block
		regOffset = 20;
	else if (this->markerIndex == 1)      // green block
		regOffset = 24;
	else if (this->markerIndex == 2)      // blue block
		regOffset = 28;
	else if (this->markerIndex == 3)      // yellow block
		regOffset = 32;
	this->xFP += this->xSpeedFP;
	this->yFP += this->ySpeedFP;
	if (this->ySpeedFP < 0) {
		if (this->yFP < this->yDestFP) {
			this->yFP = this->yDestFP;
			finished = true;
		}
	}
	else if (this->ySpeedFP > 0) {
		if (this->yFP > this->yDestFP) {
			this->yFP = this->yDestFP;
			finished = true;
		}
	}
	else if (this->xSpeedFP < 0) {
		if (this->xFP < this->xDestFP) {
			this->xFP = this->xDestFP;
			finished = true;
		}
	}
	else if (this->xSpeedFP > 0) {
		if (this->xFP > this->xDestFP) {
			this->xFP = this->xDestFP;
			finished = true;
		}
	}
	OAM[regOffset    ] = ((this->yFP >> 16) & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square
	OAM[regOffset + 1] = ((this->xFP >> 16) & 0x01FF) | (((uint16_t) 1) << 14);   // size 16x16
}


bool ScreenBoard::levelFinished() {
	GameStatus *gs = static_cast<GameStatus *>(this->interScreenData);
	uint16_t *p = gs->currentLevel.boardTiles;
	uint16_t n = Level::BOARD_SIZE;
	while (n > 0) {
		if ((((*p) & 0x0070) >> 4) != ((*p) & 0x0007))
			return false;
		n--;
		p++;
	}
	return true;
}


void ScreenBoard::commitMovement() {
	GameStatus *gs = static_cast<GameStatus *>(this->interScreenData);
	// remove block from (this->xBoardSrc, this->yBoardSrc) on gs->currentLevel
	gs->currentLevel.boardTiles[this->xBoardSrc + (this->yBoardSrc * Level::BOARD_WIDTH)] &= 0xF8;
	// put block on (this->xBoardDest, this->yBoardDest) on gs->currentLevel
	uint16_t t = gs->currentLevel.boardTiles[this->xBoardDest + (this->yBoardDest * Level::BOARD_WIDTH)];
	gs->currentLevel.boardTiles[this->xBoardDest + (this->yBoardDest * Level::BOARD_WIDTH)] = (t & 0xF8) | (this->markerIndex + 1);
	// update block location on data scructure
	this->blockLocations[this->markerIndex].x = this->xBoardDest;
	this->blockLocations[this->markerIndex].y = this->yBoardDest;
	// update marker location
	this->updateMarkerFromIndex();
}


void ScreenBoard::moveMarker(Movement m) {
	Direction &markerLocation = this->blockLocations[this->markerIndex];
	int16_t distance = 0;
	int16_t bestDistance = 10000;
	int16_t bestBlock = -1;
	for (int16_t block = 0; block < NUM_BLOCKS; block++) {
		if (!this->blockEnabled[block])
			continue;
		if (m == Movement::UP)
			distance = markerLocation.y - this->blockLocations[block].y;
		else if (m == Movement::DOWN)
			distance = this->blockLocations[block].y - markerLocation.y;
		else if (m == Movement::LEFT)
			distance = markerLocation.x - this->blockLocations[block].x;
		else if (m == Movement::RIGHT)
			distance = this->blockLocations[block].x - markerLocation.x;
		if ((distance > 0) && (distance < bestDistance)) {
			bestDistance = distance;
			bestBlock = block;
		}
	}
	if (bestBlock >= 0) {
		this->markerIndex = bestBlock;
		this->updateMarkerFromIndex();
	}
}


static char levelText[] = "Level ??";
static const uint16_t LEVEL_TEXT_SHORT_NAME_INDEX = 6;


void ScreenBoard::drawText(const char *s, int16_t x, int16_t y, uint16_t *destBG) {
	uint16_t *p = destBG + (((int32_t) y) << 5) + x;
	while (*s != 0) {
		*p = (uint16_t) (*s - ' ');
		p++;
		s++;
	}
}


void ScreenBoard::onLoad(InterScreenData *dataFromPreviousScreen) {
	this->interScreenData = dataFromPreviousScreen;
	GameStatus *gs = static_cast<GameStatus *>(this->interScreenData);
	// force blank to access VRAM
	DISPCNT = ((uint16_t) 1) << 7;
	// locate fonts at start of MODE0_VRAM (96 tiles)
	memcpy16(MODE0_VRAM, (const uint16_t *) FONT_1, FONT_SIZE >> 1);
	// locate all the 32 board tiles at position 96 (just after font, starting at tile index 96 until tile index 127)
	memcpy16(MODE0_VRAM + (96 * 32), BOARD_TILES, NUM_BOARD_TILES * 32);
	/*
	 * VRAM Layout:
	 *     BG0 for board tiles at start of VRAM                with priority 3 (lowest)
	 *     BG1 for text following BG0                          with priority 2
	 *     OBJs                                                with priority 2 (draw on top of BG1)
	 *     BG2 for ending board tiles following BG2            with priority 1
	 *     BG3 for "you lose!" or "you win!" scrolling message with priority 0 (highest)
	 */
	// draw on BG0 (and on BG2 only for endings)
	// there are 128 tiles, 128 * 64 = 8192 bytes, so map data should start at MODE0_VRAM + 8192 bytes (4096 words, 4 units of 2 Kb)
	uint16_t *dest = MODE0_VRAM + 4096;
	uint16_t *destBG2 = MODE0_VRAM + 4096 + 2048;
	memset16(destBG2, 0, 1024);    // set tile 0 on all BG2 (it is a full transparent tile)
	memset16(dest, 0, 32 * 32);   // set tile 0 by default (transparent/black)
	this->drawBoard(gs->currentLevel, dest, destBG2);
	this->drawBoardBorder2(gs->currentLevel, dest);
	// configure BG0: lowest priority (3), character base block = 0, 256 colors, screen base block = 4 (for 8192 bytes), size = 32x32 tiles = 32x32 words
	BG0CNT = ((uint16_t) 3) | (((uint16_t) 1) << 7) | (((uint16_t) 4) << 8);
	BG0HOFS = 0;    // BG0 offset (0, 0)
	BG0VOFS = 0;
	// now draw on BG1
	dest = MODE0_VRAM + 4096 + 1024;
	memset16(dest, 0, 1024);    // set tile 0 on all BG1 (it is a full transparent tile)
	levelText[LEVEL_TEXT_SHORT_NAME_INDEX] = gs->currentLevel.shortName[0];
	levelText[LEVEL_TEXT_SHORT_NAME_INDEX + 1] = gs->currentLevel.shortName[1];
	this->drawText(levelText, 20, 1, dest);
	this->drawText(gs->currentLevel.name, 20, 3, dest);
	// configure BG1: low priority (2), character base block = 0, 256 colors, screen base block = 5 (2Kb after BG0), size = 32x32 tiles = 32x32 words
	BG1CNT = ((uint16_t) 2) | (((uint16_t) 1) << 7) | (((uint16_t) 5) << 8);
	BG1HOFS = 0;    // BG1 offset (0, 0)
	BG1VOFS = 0;
	// configure BG2: high priority (1), character base block = 0, 256 colors, screen base block = 6 (2Kb after BG1), size = 32x32 tiles = 32x32 words
	// this BG2 is only for board endings (with higher priority than OBJs) because we need these endings to show OVER blocks (OBJs)
	BG2CNT = ((uint16_t) 1) | (((uint16_t) 1) << 7) | (((uint16_t) 6) << 8);
	BG2HOFS = 0;    // BG2 offset (0, 0)
	BG2VOFS = 0;
	// load OBJ palette with default palette
	memcpy16(OBJ_PALETTE, PALETTE, 256);
	// all OBJs with priority = 2 to let endings draw OVER blocks
	// marker OBJ
	OAM[2] = 512 | (((uint16_t) 2) << 10);
	// reserve OAM[4]..OAM[19] for 4 arrow objects (on "configureAndHideArrows" function member)
	// red block OBJ
	OAM[22] = 520 | (((uint16_t) 2) << 10);
	// green block OBJ
	OAM[26] = 528 | (((uint16_t) 2) << 10);
	// blue block OBJ
	OAM[30] = 536 | (((uint16_t) 2) << 10);
	// yellow block OBJ
	OAM[34] = 544 | (((uint16_t) 2) << 10);
	// configure OBJ tile data (64 bytes per tile, 32 words per tile, 4 tiles per block (5 blocks) + 4 additional tiles for arrows
	memcpy16((uint16_t *) OBJ_TILES_VRAM, (uint16_t *) BLOCK_TILES, 32 * ((4 * 5) + 4));
	this->locateBlocks(gs->currentLevel, this->markerIndex);
	this->updateMarkerFromIndex();
	this->status = Status::PRE_WAIT;
	this->waitCounter = 30;
	this->markTiltCounter = MARK_TILT_COUNTER_MAX;
	this->waitKeyCounter = 0;
	this->configureAndHideArrows();
	// timer
	if (gs->timerRunning)
		this->showRemainTime();
	else {
		// configure timer
		this->remainMinutes = gs->currentLevel.totalMinutes;
		this->remainSeconds = gs->currentLevel.totalSeconds;
		this->secondsTimer.init(*this);
		gs->timerRunning = true;
	}
	// draw BG3 (still not shown), it is for "you lose" o "you win" message
	dest = MODE0_VRAM + 4096 + 2048 + 1024;
	memset16(dest, 0, 1024);    // set tile 0 on all BG3 (it is a full transparent tile)
	// configure BG3: highest priority (0), character base block = 0, 256 colors, screen base block = 7 (2Kb after BG2), size = 32x32 tiles = 32x32 words
	// this BG2 is only for board endings (with higher priority than OBJs) because we need these endings to show OVER blocks (OBJs)
	BG3CNT = (((uint16_t) 1) << 7) | (((uint16_t) 7) << 8);
	BG3HOFS = 0;    // BG3 offset (0, SCREEN_HEIGHT), so not shown
	BG3VOFS = (uint16_t) (11 * 8);
	// set mode 0 with BG0, BG1, BG2 and OBJ layers (1-dimensional OBJ tiles) (BG3 will be enabled when board finished)
	DISPCNT = (((uint16_t) 1) << 8) | (((uint16_t) 1) << 9) | (((uint16_t) 1) << 10) | (((uint16_t) 1) << 6) | (((uint16_t) 1) << 12);
}


static const char FINISH_MESSAGE_YOU_LOSE[] = "YOU LOSE";
static const uint16_t FINISH_MESSAGE_YOU_LOSE_SIZE = 8;
static const char FINISH_MESSAGE_YOU_WIN[] = "YOU WIN";
static const uint16_t FINISH_MESSAGE_YOU_WIN_SIZE = 7;


void ScreenBoard::initFinishMessage(FinishMessage fm) {
	uint16_t *destBG3 = MODE0_VRAM + 4096 + 1024 + 1024 + 1024;
	int16_t y = 10;
	int16_t x = 10 - (((fm == FinishMessage::YOU_WIN) ? FINISH_MESSAGE_YOU_WIN_SIZE : FINISH_MESSAGE_YOU_LOSE_SIZE) >> 1);  // center of board, not center of screen
	const char *s = (fm == FinishMessage::YOU_WIN) ? FINISH_MESSAGE_YOU_WIN : FINISH_MESSAGE_YOU_LOSE;
	this->drawText(s, x, y, destBG3);
	// enable BG3
	DISPCNT |= (((uint16_t) 1) << 11);
}


void ScreenBoard::hideArrows() {
	// up arrow
	OAM[4] = (SCREEN_HEIGHT & 0x00FF) | (((uint16_t) 1) << 13) | (((uint16_t) 1) << 9);    // 256 colors, square, hidden
	OAM[5] = (SCREEN_WIDTH & 0x01FF);                                                      // size 8x8
	// down arrow
	OAM[8] = (SCREEN_HEIGHT & 0x00FF) | (((uint16_t) 1) << 13) | (((uint16_t) 1) << 9);    // 256 colors, square, hidden
	OAM[9] = (SCREEN_WIDTH & 0x01FF);                                                      // size 8x8
	// left arrow
	OAM[12] = (SCREEN_HEIGHT & 0x00FF) | (((uint16_t) 1) << 13) | (((uint16_t) 1) << 9);   // 256 colors, square, hidden
	OAM[13] = (SCREEN_WIDTH & 0x01FF);                                                     // size 8x8
	// right arrow
	OAM[16] = (SCREEN_HEIGHT & 0x00FF) | (((uint16_t) 1) << 13) | (((uint16_t) 1) << 9);   // 256 colors, square, hidden
	OAM[17] = (SCREEN_WIDTH & 0x01FF);                                                     // size 8x8
}


void ScreenBoard::configureAndHideArrows() {
	// arrows also with priority = 2 (same as rest of OBJs)
	// up arrow
	OAM[6] = 552 | (((uint16_t) 2) << 10);
	// down arrow
	OAM[10] = 554 | (((uint16_t) 2) << 10);
	// left arrow
	OAM[14] = 556 | (((uint16_t) 2) << 10);
	// right arrow
	OAM[18] = 558 | (((uint16_t) 2) << 10);
	this->hideArrows();
}


void ScreenBoard::showRemainTime() {
	const int16_t X_TIMER = 22;
	const int16_t Y_TIMER = 6;
	const uint16_t TIMER_BG_OFFSET = (Y_TIMER * 32) + X_TIMER;
	uint16_t *p = MODE0_VRAM + 4096 + 1024 + TIMER_BG_OFFSET;     // BG1 map
	*p = (this->remainMinutes / 10) + '0' - ' ';
	*(p + 1) = (this->remainMinutes % 10) + '0' - ' ';
	*(p + 2) = ':' - ' ';
	*(p + 3) = (this->remainSeconds / 10) + '0' - ' ';
	*(p + 4) = (this->remainSeconds % 10) + '0' - ' ';
}


void ScreenBoard::oneSecondElapsed(SecondsTimer &timer) {
	//const int16_t X_TIMER = 22;
	//const int16_t Y_TIMER = 3;
	//const uint16_t TIMER_BG_OFFSET = (Y_TIMER * 32) + X_TIMER;
	//uint16_t *p = MODE0_VRAM + 4096 + 1024 + TIMER_BG_OFFSET;     // BG1 map
	if (this->remainSeconds == 0) {
		if (this->remainMinutes > 0) {
			this->remainMinutes--;
			this->remainSeconds = 59;
		}
	}
	else
		this->remainSeconds--;
	this->showRemainTime();
	/*
	*p = (this->remainMinutes / 10) + '0' - ' ';
	*(p + 1) = (this->remainMinutes % 10) + '0' - ' ';
	*(p + 2) = ':' - ' ';
	*(p + 3) = (this->remainSeconds / 10) + '0' - ' ';
	*(p + 4) = (this->remainSeconds % 10) + '0' - ' ';
	*/
}


void ScreenBoard::updateArrowsMask() {
	GameStatus *gs = static_cast<GameStatus *>(this->interScreenData);
	Direction &marker = this->blockLocations[this->markerIndex];
	this->arrowsMask = 0;
	if (this->isBoardNeighbourPassable(gs->currentLevel, marker.x, marker.y, DIRECTIONS[DIRECTION_TOP], ConsiderTheBlocks::YES))
		this->arrowsMask |= ARROW_UP_MASK;
	if (this->isBoardNeighbourPassable(gs->currentLevel, marker.x, marker.y, DIRECTIONS[DIRECTION_BOTTOM], ConsiderTheBlocks::YES))
		this->arrowsMask |= ARROW_DOWN_MASK;
	if (this->isBoardNeighbourPassable(gs->currentLevel, marker.x, marker.y, DIRECTIONS[DIRECTION_LEFT], ConsiderTheBlocks::YES))
		this->arrowsMask |= ARROW_LEFT_MASK;
	if (this->isBoardNeighbourPassable(gs->currentLevel, marker.x, marker.y, DIRECTIONS[DIRECTION_RIGHT], ConsiderTheBlocks::YES))
		this->arrowsMask |= ARROW_RIGHT_MASK;
}


void ScreenBoard::showArrowsFromArrowsMask() {
	Direction &marker = this->blockLocations[this->markerIndex];
	int16_t x = 8 + (marker.x << 4);
	int16_t y = 8 + (marker.y << 4);
	if (this->arrowsMask & ARROW_UP_MASK) {
		// show UP arrow
		OAM[4] = ((y - 8) & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square, shown
		OAM[5] = ((x + 4) & 0x01FF);                            // size 8x8
	}
	else {
		// hide UP arrow
		OAM[4] |= (((uint16_t) 1) << 9);
	}
	if (this->arrowsMask & ARROW_DOWN_MASK) {
		// show DOWN arrow
		OAM[8] = ((y + 16) & 0x00FF) | (((uint16_t) 1) << 13);    // 256 colors, square, shown
		OAM[9] = ((x + 4) & 0x01FF);                              // size 8x8
	}
	else {
		// hide DOWN arrow
		OAM[8] |= (((uint16_t) 1) << 9);
	}
	if (this->arrowsMask & ARROW_LEFT_MASK) {
		// show LEFT arrow
		OAM[12] = ((y + 4) & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square, shown
		OAM[13] = ((x - 8) & 0x01FF);                            // size 8x8
	}
	else {
		// hide LEFT arrow
		OAM[12] |= (((uint16_t) 1) << 9);
	}
	if (this->arrowsMask & ARROW_RIGHT_MASK) {
		// show RIGHT arrow
		OAM[16] = ((y + 4) & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square, hidden
		OAM[17] = ((x + 16) & 0x01FF);                             // size 8x8
	}
	else {
		// hide RIGHT arrow
		OAM[16] |= (((uint16_t) 1) << 9);
	}
}


void ScreenBoard::showArrowsAndUpdateArrowsMask() {
	GameStatus *gs = static_cast<GameStatus *>(this->interScreenData);
	Direction &marker = this->blockLocations[this->markerIndex];
	int16_t x = 8 + (marker.x << 4);
	int16_t y = 8 + (marker.y << 4);
	this->arrowsMask = 0;
	if (this->isBoardNeighbourPassable(gs->currentLevel, marker.x, marker.y, DIRECTIONS[DIRECTION_TOP], ConsiderTheBlocks::YES)) {
		// show UP arrow
		OAM[4] = ((y - 8) & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square, shown
		OAM[5] = ((x + 4) & 0x01FF);                            // size 8x8
		this->arrowsMask |= ARROW_UP_MASK;
	}
	else {
		// hide UP arrow
		OAM[4] |= (((uint16_t) 1) << 9);
	}
	if (this->isBoardNeighbourPassable(gs->currentLevel, marker.x, marker.y, DIRECTIONS[DIRECTION_BOTTOM], ConsiderTheBlocks::YES)) {
		// show DOWN arrow
		OAM[8] = ((y + 16) & 0x00FF) | (((uint16_t) 1) << 13);    // 256 colors, square, shown
		OAM[9] = ((x + 4) & 0x01FF);                              // size 8x8
		this->arrowsMask |= ARROW_DOWN_MASK;
	}
	else {
		// hide DOWN arrow
		OAM[8] |= (((uint16_t) 1) << 9);
	}
	if (this->isBoardNeighbourPassable(gs->currentLevel, marker.x, marker.y, DIRECTIONS[DIRECTION_LEFT], ConsiderTheBlocks::YES)) {
		// show LEFT arrow
		OAM[12] = ((y + 4) & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square, shown
		OAM[13] = ((x - 8) & 0x01FF);                            // size 8x8
		this->arrowsMask |= ARROW_LEFT_MASK;
	}
	else {
		// hide LEFT arrow
		OAM[12] |= (((uint16_t) 1) << 9);
	}
	if (this->isBoardNeighbourPassable(gs->currentLevel, marker.x, marker.y, DIRECTIONS[DIRECTION_RIGHT], ConsiderTheBlocks::YES)) {
		// show RIGHT arrow
		OAM[16] = ((y + 4) & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square, hidden
		OAM[17] = ((x + 16) & 0x01FF);                             // size 8x8
		this->arrowsMask |= ARROW_RIGHT_MASK;
	}
	else {
		// hide RIGHT arrow
		OAM[16] |= (((uint16_t) 1) << 9);
	}
}


void ScreenBoard::saveLevelAsFinished() {
	GameStatus *gs = static_cast<GameStatus *>(this->interScreenData);
	gs->saveFile.levels[gs->currentLevelIndex] = SaveFile::LevelStatus::FINISHED;
	// check whether we must to unlock other levels
	for (uint8_t i = 0; i < Level::MAX_LEVELS_TO_UNLOCK; i++) {
		uint8_t index = gs->currentLevel.levelsToUnlock[i];
		if (index != Level::NULL_LEVEL) {
			if (gs->saveFile.levels[index] == SaveFile::LevelStatus::LOCKED)
				gs->saveFile.levels[index] = SaveFile::LevelStatus::UNLOCKED;
		}
	}
	SaveFileManager sfm;
	sfm.save(gs->saveFile.position, gs->saveFile);
}


Screen *ScreenBoard::onVBlank(Keypad &keypad) {
	uint16_t keys = 0;
	if (!keypad.keyQueue.empty()) {
		keys = keypad.keyQueue.head();
		keypad.keyQueue.pop();
	}
	this->secondsTimer.onVBlank();
	if (this->status == Status::PRE_WAIT) {
		this->waitCounter--;
		if (this->waitCounter == 0)
			this->status = Status::MARK;
	}
	if (this->status == Status::MARK) {
		this->markTiltCounter--;
		if (this->markTiltCounter == 0) {
			this->markTiltCounter = MARK_TILT_COUNTER_MAX;
			// show marker
			OAM[0] = OAM[0] & ~(((uint16_t) 1) << 9);
		}
		if (this->markTiltCounter == MARK_TILT_COUNTER_HIDE) {
			// hide marker
			OAM[0] = OAM[0] | (((uint16_t) 1) << 9);
		}
		if (keys & KEY_UP)
			this->moveMarker(Movement::UP);
		else if (keys & KEY_DOWN)
			this->moveMarker(Movement::DOWN);
		else if (keys & KEY_LEFT)
			this->moveMarker(Movement::LEFT);
		else if (keys & KEY_RIGHT)
			this->moveMarker(Movement::RIGHT);
		else if (keys & KEY_A) {
			this->updateArrowsMask();
			if (this->arrowsMask != 0) {
				// hide marker
				OAM[0] = OAM[0] | (((uint16_t) 1) << 9);
				// show arrows
				//this->showArrowsAndUpdateArrowsMask();
				//this->updateArrowsMask();
				this->showArrowsFromArrowsMask();
				this->status = Status::STEER;
			}
		}
		else if (keys & KEY_START)
			return this->pauseScreen;
		if ((this->remainMinutes == 0) && (this->remainSeconds == 0))
			this->status = Status::END_LOSE;
	}
	else if (this->status == Status::STEER) {
		this->markTiltCounter--;
		if (this->markTiltCounter == 0) {
			this->markTiltCounter = MARK_TILT_COUNTER_MAX;
			// show marker
			OAM[0] = OAM[0] & ~(((uint16_t) 1) << 9);
		}
		if (this->markTiltCounter == MARK_TILT_COUNTER_HIDE) {
			// hide marker
			OAM[0] = OAM[0] | (((uint16_t) 1) << 9);
		}
		if ((keys & KEY_UP) && (this->arrowsMask & ARROW_UP_MASK)) {
			this->startMoveBlockOnMarker(Movement::UP);
			this->status = Status::MOVE;
		}
		else if ((keys & KEY_DOWN) && (this->arrowsMask & ARROW_DOWN_MASK)) {
			this->startMoveBlockOnMarker(Movement::DOWN);
			this->status = Status::MOVE;
		}
		else if ((keys & KEY_LEFT) && (this->arrowsMask & ARROW_LEFT_MASK)) {
			this->startMoveBlockOnMarker(Movement::LEFT);
			this->status = Status::MOVE;
		}
		else if ((keys & KEY_RIGHT) && (this->arrowsMask & ARROW_RIGHT_MASK)) {
			this->startMoveBlockOnMarker(Movement::RIGHT);
			this->status = Status::MOVE;
		}
		else if (keys & KEY_B) {
			this->hideArrows();
			this->status = Status::MARK;
		}
		if ((this->remainMinutes == 0) && (this->remainSeconds == 0))
			this->status = Status::END_LOSE;
	}
	else if (this->status == Status::MOVE) {
		bool finished = false;
		this->updateMoveBlockOnMarker(finished);
		if (finished) {
			this->commitMovement();
			if (this->levelFinished())
				this->status = Status::END_WIN;
			else
				this->status = Status::MARK;
		}
	}
	else if (this->status == Status::END_WIN) {
		this->saveLevelAsFinished();
		this->secondsTimer.stop();
		this->initFinishMessage(FinishMessage::YOU_WIN);
		this->status = Status::ENDING;
		this->endingMessageYFP = ((int32_t) (11 * 8)) << 16;
	}
	else if (this->status == Status::END_LOSE) {
		this->initFinishMessage(FinishMessage::YOU_LOSE);
		this->status = Status::ENDING;
		this->endingMessageYFP = ((int32_t) (11 * 8)) << 16;
	}
	else if (this->status == Status::ENDING) {
		this->endingMessageYFP -= ENDING_MESSAGE_SPEED_FP;
		if (this->endingMessageYFP < 0) {
			this->endingMessageYFP = 0;
			this->waitCounter = WAIT_COUNTER_AFTER_FINISH;
			this->status = Status::FINISHED;
		}
		BG3VOFS = (this->endingMessageYFP >> 16);
	}
	else if (this->status == Status::FINISHED) {
		this->waitCounter--;
		if (this->waitCounter == 0)
			return this->afterFinishScreen;
	}
	return this;
}


InterScreenData *ScreenBoard::onUnload() {
	this->hideSprites();
	this->hideArrows();
	return this->interScreenData;
}


void ScreenBoard::hideSprites() {
	// locate marker outsize screen
	OAM[0] = (SCREEN_HEIGHT & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square
	OAM[1] = (SCREEN_WIDTH & 0x01FF) | (((uint16_t) 1) << 14);    // size 16x16
	// locate red block outsize screen
	OAM[20] = (SCREEN_HEIGHT & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square
	OAM[21] = (SCREEN_WIDTH & 0x01FF) | (((uint16_t) 1) << 14);    // size 16x16
	// locate green block outsize screen
	OAM[24] = (SCREEN_HEIGHT & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square
	OAM[25] = (SCREEN_WIDTH & 0x01FF) | (((uint16_t) 1) << 14);    // size 16x16
	// locate blue block outsize screen
	OAM[28] = (SCREEN_HEIGHT & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square
	OAM[29] = (SCREEN_WIDTH & 0x01FF) | (((uint16_t) 1) << 14);    // size 16x16
	// locate yellow block outsize screen
	OAM[32] = (SCREEN_HEIGHT & 0x00FF) | (((uint16_t) 1) << 13);   // 256 colors, square
	OAM[33] = (SCREEN_WIDTH & 0x01FF) | (((uint16_t) 1) << 14);    // size 16x16
}


