

#include "ScreenMenu.H"
#include "gba.H"
#include "fonts.H"
#include "palette.H"
#include "tiles.H"


using namespace std;
using namespace avelino;


int16_t ScreenMenu::drawText(int16_t x, int16_t y, const char *s) {
	if (x < 0)
		x = 15 - ((int16_t) (strlen(s) >> 1));   // 30 tiles width, so 15 tiles half-width
	uint16_t *dest = (MODE0_VRAM + 4096 + 1024) + (y << 5) + x;   // (y * 32) + x
	while (*s != 0) {
		*dest = (uint16_t) (*s - ' ');
		s++;
		dest++;
	}
	return x;
}


void ScreenMenu::locateMarker(int16_t x, int16_t y) {
	OAM[0] = (y & 0x00FF) | (((uint16_t) 1) << 13);
	OAM[1] = x & 0x01FF;
}


void ScreenMenu::onLoad(InterScreenData *dataFromPreviousScreen) {
	this->interScreenData = dataFromPreviousScreen;
	// force blank to access VRAM
	DISPCNT = ((uint16_t) 1) << 7;
	// locate fonts at start of MODE0_VRAM (96 tiles)
	memcpy16(MODE0_VRAM, (const uint16_t *) FONT_1, FONT_SIZE >> 1);
	// generate aditional flat tiles (indices 96, 97, 98 and 99)
	uint16_t *dest = MODE0_VRAM + (FONT_SIZE >> 1);
	const uint16_t NUM_COLORS = 4;
	const uint16_t COLORS[NUM_COLORS] = { 0x1313, 0x2323, 0x3333, 0x4343 };
	const uint16_t *src = COLORS;
	for (uint16_t i = 0; i < NUM_COLORS; i++, dest += 32, src++)
		memset16(dest, *src, 32);
	// there are 100 tiles, 100 * 64 = 6400 bytes, so map data should start at MODE0_VRAM + 8192 bytes (units of 2 Kb)
	dest = MODE0_VRAM + 4096;
	// fill pattern (32x32 tiles, 256x256 pixels)
	for (uint16_t y = 0; y < 32; y++) {
		uint16_t i = y & 0x0003;
		for (uint16_t x = 0; x < 32; x++, dest++, i++)
			*dest = 96 + (i & 0x0003);
	}
	// load BG palette
	memcpy16(BG_PALETTE, PALETTE, 256);
	// configure BG0: lowest priority (3), character base block = 0, 256 colors, screen base block = 4 (for 8192 bytes), size = 32x32 tiles = 32x32 words
	BG0CNT = ((uint16_t) 3) | (((uint16_t) 1) << 7) | (((uint16_t) 4) << 8);
	this->bg0Offset = 0;
	BG0HOFS = (this->bg0Offset >> 16) & 0x01FF;
	BG0VOFS = (this->bg0Offset >> 16) & 0x01FF;
	// draw text
	dest = MODE0_VRAM + 4096 + 1024;
	memset16(dest, 0, 1024);    // set tile 0 on all BG1 (it is a full transparent tile)
	const char *text;
	uint16_t i = 0;
	int16_t y;
	while ((text = this->getTitleText(i, y)) != NULL) {
		this->drawText(-1, y, text);
		i++;
	}
	i = 0;
	this->optionsX = SCREEN_WIDTH;
	int16_t *optionsYPtr = this->optionsY;
	while ((text = this->getOptionText(i, y)) != NULL) {
		int16_t realX = this->drawText(-1, y, text);
		realX <<= 3;           // tile coord --> screen coord
		if (realX < this->optionsX)
			this->optionsX = realX;
		*optionsYPtr = y << 3;
		optionsYPtr++;
		i++;
	}
	this->numOptions = i;
	this->currentOptionIndex = 0;
	this->markerYFP = ((int32_t) this->optionsY[0]) << 16;
	this->minMarkerXFP = (((int32_t) this->optionsX - 10) << 16) - MARKER_X_WIDTH_FP;
	this->markerXFP = this->minMarkerXFP;
	this->maxMarkerXFP = this->minMarkerXFP + MARKER_X_WIDTH_FP;
	this->tiltManager.init(this->minMarkerXFP, this->maxMarkerXFP, MARKER_X_SPEED_FP, this);
	if ((text = this->getBottomText()) != NULL)
		this->drawText(-1, 19, text);
	// configure BG1: low priority (2), character base block = 0, 256 colors, screen base block = 5 (2Kb after BG0), size = 32x32 tiles = 32x32 words
	BG1CNT = ((uint16_t) 2) | (((uint16_t) 1) << 7) | (((uint16_t) 5) << 8);
	// configure OBJ palette for simple arrow marker
	OBJ_PALETTE[0] = 0;
	OBJ_PALETTE[1] = RGB555(0, 0, 15);
	OBJ_PALETTE[2] = RGB555(0, 30, 0);
	// configure OBJ tile data for simple arrow marker
	memcpy16((uint16_t *) OBJ_TILES_VRAM, OPTION_MARKER_OBJ_TILE, 32);
	// locate OBJ (arrow marker)
	OAM[2] = 512;
	//this->locateMarker(this->optionsX - 10, this->optionsY[0]);
	this->locateMarker(this->markerXFP >> 16, this->markerYFP >> 16);
	this->waitCounter = INITIAL_WAIT_COUNTER_MAX;
	this->markerStatus = MarkerStatus::WAIT;
	// set mode 0 with both BG0 and BG1 and enable OBJ
	DISPCNT = (((uint16_t) 1) << 8) | (((uint16_t) 1) << 9) | (((uint16_t) 1) << 12);
}


void ScreenMenu::updateTilt(TiltManager &tiltManager, int32_t v) {
	this->markerXFP = v;
}


Screen *ScreenMenu::onVBlank(Keypad &keypad) {
	// update BG0 scroll
	this->bg0Offset += BG0_OFFSET_INC;
	BG0HOFS = (this->bg0Offset >> 16) & 0x01FF;
	BG0VOFS = (this->bg0Offset >> 16) & 0x01FF;
	// update marker tilt
	this->tiltManager.onVBlank();
	// update marker position
	if (this->markerStatus == MarkerStatus::STOPED) {
		if (keypad.mask & KEY_B)
			return this->backSelected();
		else if (keypad.mask & KEY_A)
			return this->optionSelected(this->currentOptionIndex);
		else if ((keypad.mask & KEY_DOWN) && (this->currentOptionIndex < (this->numOptions - 1)))
			this->markerStatus = MarkerStatus::MOVING_DOWN;
		else if ((keypad.mask & KEY_UP) && (this->currentOptionIndex > 0))
			this->markerStatus = MarkerStatus::MOVING_UP;
	}
	else if (this->markerStatus == MarkerStatus::MOVING_DOWN) {
		this->markerYFP += MARKER_Y_SPEED_FP;
		if ((this->markerYFP >> 16) >= this->optionsY[this->currentOptionIndex + 1]) {
			this->currentOptionIndex++;
			this->markerYFP = ((int32_t) this->optionsY[this->currentOptionIndex]) << 16;
			this->waitCounter = AFTER_KEY_WAIT_COUNTER_MAX;
			this->markerStatus = MarkerStatus::WAIT;
		}
	}
	else if (this->markerStatus == MarkerStatus::MOVING_UP) {
		this->markerYFP -= MARKER_Y_SPEED_FP;
		if ((this->markerYFP >> 16) <= this->optionsY[this->currentOptionIndex - 1]) {
			this->currentOptionIndex--;
			this->markerYFP = ((int32_t) this->optionsY[this->currentOptionIndex]) << 16;
			this->waitCounter = AFTER_KEY_WAIT_COUNTER_MAX;
			this->markerStatus = MarkerStatus::WAIT;
		}
	}
	else if (this->markerStatus == MarkerStatus::WAIT) {
		this->waitCounter--;
		if (this->waitCounter == 0)
			this->markerStatus = MarkerStatus::STOPED;
	}
	// update marker position
	this->locateMarker(this->markerXFP >> 16, this->markerYFP >> 16);
	return this;
}


InterScreenData *ScreenMenu::onUnload() {
	this->locateMarker(SCREEN_WIDTH, SCREEN_HEIGHT);
	BG0HOFS = 0;
	BG0VOFS = 0;
	return this->interScreenData;
}

