

#include <iostream>
#include <fstream>
#include <iomanip>
#include <time.h>
#include <stdlib.h>
#include <vector>
#include <string.h>


using namespace std;


class Vector {
	public:
		int x;
		int y;
		Vector(int x, int y) : x(x), y(y) { };
};


ostream &operator << (ostream &os, const Vector &v) {
	os << dec << setw(0) << "(" << v.x << ", " << v.y << ")";
	return os;
}


class Block : public Vector {
	public:
		int number;
		Block(int number, int x, int y) : number(number), Vector(x, y) { };
};


ostream &operator << (ostream &os, const Block &block) {
	os << dec << setw(0) << "(" << block.number << ": " << block.x << ", " << block.y << ")";
	return os;
}


class Direction : public Vector {
	public:
		int oppositeIndex;
		Direction(int x, int y, int oppositeIndex) : Vector(x, y), oppositeIndex(oppositeIndex) { };
};


class Level {
	public:
		static const uint16_t MAX_BLOCKS = 4;
		static const uint16_t BOARD_WIDTH = 9;
		static const uint16_t BOARD_HEIGHT = 9;
		static const uint16_t BOARD_SIZE = BOARD_WIDTH * BOARD_HEIGHT;
		uint16_t boardTiles[BOARD_SIZE];
		// each board tile:    15  14  13  12  11 10  9  8  7  6  5  4  3  2  1  0
		//                     |                         |  P  D2 D1 D0    B0 B1 B0
		//                     +------------+------------+  |     |           |
		//                                  |               |     |           +--- block here (1..7) or empty tile (0)
		//                                  |               |     +--------------- finish for block (1..7) or empty tile (0)
		//                                  |               +--------------------- 0 is passable, 1 is not passable
		//                                  +------------------------------------- reserved (for special features)
		static const int NUM_DIRECTIONS = 4;
		static const Direction DIRECTIONS[NUM_DIRECTIONS];
		vector<Block> blocks;
		Level() { };
		Level(const Level &other);
		void load(istream &is);
		void save(ostream &os);
		void randomizeBlockEndings(int blockMask);
		bool isPassable(int x, int y);
		bool backIterate();
		void backMove(int blockIndex, const Vector &dir);
};


Level::Level(const Level &other) {
	memcpy(this->boardTiles, other.boardTiles, BOARD_SIZE * sizeof(uint16_t));
	for (int y = 0; y < BOARD_HEIGHT; y++) {
		for (int x = 0; x < BOARD_WIDTH; x++) {
			int offset = (y * BOARD_WIDTH) + x;
			uint16_t v = this->boardTiles[offset];
			if (v & 0x07)
				this->blocks.push_back({v + 1, x, y});
		}
	}
}


const Direction Level::DIRECTIONS[NUM_DIRECTIONS] = {
	{-1, 0, 1}, {1, 0, 0}, {0, -1, 3}, {0, 1, 2}
};


void Level::load(istream &is) {
	is.read((char *) this->boardTiles, BOARD_SIZE * sizeof(uint16_t));
}


void Level::save(ostream &os) {
	os.write((const char *) this->boardTiles, BOARD_SIZE * sizeof(uint16_t));
}


bool Level::isPassable(int x, int y) {
	if ((x < 0) || (x >= BOARD_WIDTH) || (y < 0) || (y >= BOARD_HEIGHT))
		return false;
	int offset = (y * BOARD_WIDTH) + x;
	return ((this->boardTiles[offset] & 0x0080) || (this->boardTiles[offset] & 0x0007)) ? false : true;
}


void Level::randomizeBlockEndings(int blockMask) {
	for (int i = 0; i < Level::MAX_BLOCKS; i++) {
		int mask = 1 << i;
		if (blockMask & mask) {
			int x, y, offset;
			do {
				x = rand() % BOARD_WIDTH;
				y = rand() % BOARD_HEIGHT;
				offset = (y * BOARD_WIDTH) + x;
			} while (this->boardTiles[offset] != 0);
			this->boardTiles[offset] = ((i + 1) << 4) | (i + 1);   // locate end and block in the same tile
			this->blocks.push_back({i + 1, x, y});
		}
	}
}


bool Level::backIterate() {
	cout << "backIterate" << endl;
	int blockIndex;
	bool availableDirection[NUM_DIRECTIONS] = {false, false, false, false};
	bool someAvailableDirection;
	bool availableBlock[this->blocks.size()];
	for (int i = 0; i < this->blocks.size(); i++)
		availableBlock[i] = false;
	int acc;
	do {
		blockIndex = rand() % this->blocks.size();
		availableBlock[blockIndex] = true;
		cout << "    checking block number " << (blockIndex + 1) << endl;
		someAvailableDirection = false;
		for (int i = 0; i < NUM_DIRECTIONS; i++) {
			int x = this->blocks[blockIndex].x + DIRECTIONS[i].x;
			int y = this->blocks[blockIndex].y + DIRECTIONS[i].y;
			bool p = this->isPassable(x, y);
			cout << "        checking direction " << DIRECTIONS[i] << " so checking board coord: (" << x << ", " << y << ")... passable=" << p << endl;
			if (!p) {
				cout << "            then can be back-moved to " << DIRECTIONS[DIRECTIONS[i].oppositeIndex] << endl;
				availableDirection[DIRECTIONS[i].oppositeIndex] = true;
				someAvailableDirection = true;
			}
			else {
				cout << "            then can NO be back-moved to " << DIRECTIONS[DIRECTIONS[i].oppositeIndex] << endl;
				availableDirection[DIRECTIONS[i].oppositeIndex] = false;
			}
		}
		acc = 0;
		for (int i = 0; i < this->blocks.size(); i++)
			acc += (availableBlock[i]) ? 1 : 0;
	} while (!someAvailableDirection && (acc < this->blocks.size()));
	if (someAvailableDirection) {
		// at this point we have a valid blockIndex with a non empty availableDirection set
		cout << dec << setw(0) << "    Block " << (blockIndex + 1) << " can be back-moved to: ";
		for (int i = 0; i < NUM_DIRECTIONS; i++) {
			if (availableDirection[i])
				cout << DIRECTIONS[i] << " ";
		}
		cout << "." << endl;
		// choose a random direction from available for back-move
		int direction;
		do {
			direction = rand() % NUM_DIRECTIONS;
		} while (!availableDirection[direction]);
		// move blockIndex in that direction until touch a wall of another block
		cout << "    moving block " << (blockIndex + 1) << " to direction " << DIRECTIONS[direction] << endl;
		this->backMove(blockIndex, DIRECTIONS[direction]);
		return true;
	}
	else {
		cout << "    no directions available" << endl;
		return false;
	}
}


void Level::backMove(int blockIndex, const Vector &dir) {
	int x = this->blocks[blockIndex].x;
	int y = this->blocks[blockIndex].y;
	int offset = (y * BOARD_WIDTH) + x;
	this->boardTiles[offset] = (this->boardTiles[offset] & 0x00F0);   // remove blockIndex from current location
	int newX = x;
	int newY = y;
	do {
		x = newX;
		y = newY;
		newX += dir.x;
		newY += dir.y;
	} while (this->isPassable(newX, newY));
	offset = (y * BOARD_WIDTH) + x;
	this->boardTiles[offset] = (this->boardTiles[offset] & 0x00F0) | (blockIndex + 1);    // put blockIndex at new location
	this->blocks[blockIndex].x = x;
	this->blocks[blockIndex].y = y;
}


ostream &operator << (ostream &os, const Level &level) {
	for (uint16_t y = 0; y < Level::BOARD_HEIGHT; y++) {
		for (uint16_t x = 0; x < Level::BOARD_HEIGHT; x++) {
			uint16_t v = level.boardTiles[(y * Level::BOARD_WIDTH) + x];
			if (v == 0)
				cout << "  ____, ";
			else
				cout << "0x" << hex << setw(4) << setfill('0') << v << ", ";
		}
		cout << endl;
	}
	return os;
}


int main(int argc, char *argv[]) {
	if (argc < 4) {
		cerr << "Syntax: " << argv[0] << " template.bin blockMask maxIterations" << endl;
		return 1;
	}
	srand((unsigned) time(NULL));
	ifstream ifs(argv[1]);
	Level level;
	level.load(ifs);
	level.randomizeBlockEndings(atoi(argv[2]));
	int n = atoi(argv[3]);
	do {
		cout << level;
		n--;
	} while (level.backIterate() && (n > 0));
	return 0;
}

